/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/broker.cc
 *  
 *  Description:
 *    <--->
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#include <map>
#include <string>

#include "broker.h"
#include "io/io.h"
#include "operators.h"
#include "parsers/imap/recursivedescent/recursivedescent.h"
#include "session.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
BrokerFactory::BrokerFactory(void)
{
  brokers[Session::NONAUTHENTICATED] = new Broker();
  brokers[Session::AUTHENTICATED] = new Broker();
  brokers[Session::SELECTED] = new Broker();
}

//----------------------------------------------------------------------
BrokerFactory::~BrokerFactory(void)
{
  //  for (map<int, Broker *>::iterator i = brokers.begin(); 
  //       i != brokers.end(); ++i)
  //    delete i->second;
}

//----------------------------------------------------------------------
BrokerFactory &BrokerFactory::getInstance(void)
{
  static BrokerFactory brokerfactory;
  return brokerfactory;
}

//----------------------------------------------------------------------
void BrokerFactory::addCapability(const std::string &c)
{
  for (map<int, Broker *>::iterator i = brokers.begin(); 
       i != brokers.end(); ++i) {
    CapabilityOperator * o;
    o = dynamic_cast<CapabilityOperator*>(i->second->get("CAPABILITY"));
    if (o != 0) {
      o->addCapability(c);
      break;
    }
  }
}

//----------------------------------------------------------------------
void BrokerFactory::assign(Operator *o)
{
  int deletable = true;
  for (map<int, Broker *>::iterator i = brokers.begin(); 
       i != brokers.end(); ++i)
    if (i->first & o->getState()) {
      i->second->assign(o, deletable);
      deletable = false;
    }
}

//----------------------------------------------------------------------
Operator *BrokerFactory::getOperator(int state, const string &name) const
{
  if (brokers.find(state) == brokers.end())
    return NULL;
  else
    return brokers.find(state)->second->get(name);
}

//----------------------------------------------------------------------
Broker &BrokerFactory::getBroker(int state)
{
  if (brokers.find(state) == brokers.end())
    throw CriticalException("No appropriate broker for state.");

  return *brokers[state];
}

//----------------------------------------------------------------------
Broker::Broker(void)
{
}

//----------------------------------------------------------------------
Broker::~Broker(void)
{
}

//----------------------------------------------------------------------
void Broker::assign(Operator *o, bool deletable)
{
  deletables[o->getName()] = deletable;
  operators[o->getName()] = o;
}

Operator * Broker::get(const string &name) const
{
  if (operators.find(name) == operators.end())
    return NULL;
  else
    return operators.find(name)->second;
}

//----------------------------------------------------------------------
bool Broker::process(Depot &depot, Command &command) 
  const throw (OperatorException &)
{
  map<string, Operator *>::const_iterator pos;
  if ((pos = operators.find(command.getName())) == operators.end())
    throw OperatorException(Operator::BAD, 
			    "command undefined or disallowed in current state");

  Operator *o = pos->second;
  o->process(depot, command);
  return true;
}

//----------------------------------------------------------------------
bool Broker::parse(Command & c_in) const
{
  string t;

  try {
    if (!expectTag(t))
      throw SyntaxException("First token should be a tag");
    
    if (!expectSPACE())
      throw SyntaxException("expected a single space");

    c_in.setTag(t);

    map<string, Operator *>::const_iterator i = operators.begin();
    while (i != operators.end()) {
      if (i->second->parse(c_in)) return true;
      ++i;
    }
  } catch (InputException &e) {
    e.addReason("in Broker::parse");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in Broker::parse");
    throw e;
  } catch (BincException &e) {
    IO &logger = IOFactory::getInstance().get(2);

    logger << "Caught BincException in Broker::parse: "
	   << e.getReason() << endl;
    throw;
  }

  throw SyntaxException(c_in.getName() 
			+ " undefined or disallowed in current state");
}
