/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/io/io.h
 *  
 *  Description:
 *    Implementation of the IO class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#ifdef HAVE_OSTREAM
#include <ostream>
#else
#include <ostream.h>
#endif

#include "../session.h"
#include "io.h"


using namespace ::std;
using namespace Binc;

//------------------------------------------------------------------------
IO::IO()
{
  pid = getpid();
  enabled = true;
  flushesOnEndl = true;
  mode = MODE_PLAIN;
  useLogPrefix = false;
  seqnr = 0;
  buffersize = 8192;
  transfertimeout = 0;
  inputsize = 0;
  inputlimit = true;
}

//------------------------------------------------------------------------
IO::IO(FILE *fp)
{
  pid = getpid();
  setFd(fp);
  enabled = true;
  flushesOnEndl = true;
  mode = MODE_PLAIN;
  useLogPrefix = false;
  seqnr = 0;
  buffersize = 8192;
  transfertimeout = 0;
  inputsize = 0;
  inputlimit = true;
}

//------------------------------------------------------------------------
IO::~IO(void)
{
  switch (mode) {
  case MODE_SYSLOG:
    closelog();
    break;
  default:
    break;
  }
}

//------------------------------------------------------------------------
void IO::enable(void)
{
  enabled = true;
}

//------------------------------------------------------------------------
void IO::enableLogPrefix(void)
{
  useLogPrefix = true;
}

//------------------------------------------------------------------------
void IO::disable(void)
{
  enabled = false;
}

//------------------------------------------------------------------------
void IO::flushContent(void)
{
  if (!enabled) return;

  string s = outputBuffer.str();
  string tmpstr;
  if (mode == MODE_SYSLOG) {
    tmpstr = "";
    for (string::iterator i = s.begin(); i != s.end(); ++i)
      if (*i != '\r' && *i != '\n') {
	tmpstr += *i;
      } else {
	writeStr(tmpstr);
	tmpstr = "";
      }

    if (tmpstr != "")
      writeStr(tmpstr);
  } else
    writeStr(s);


  outputBuffer.clear();
}

//------------------------------------------------------------------------
void IO::flushOnEndl(void)
{
  flushesOnEndl = true;
}

//------------------------------------------------------------------------
void IO::noFlushOnEndl(void)
{
  flushesOnEndl = false;
}

//------------------------------------------------------------------------
void IO::clear(void)
{
  if (!enabled) return;

  outputBuffer.clear();
}

//------------------------------------------------------------------------
void IO::setModeSyslog(const string &servicename, int facility = LOG_DAEMON)
{
  if (mode != MODE_SYSLOG) {
    openlog(servicename.c_str(), LOG_PID, facility);
    mode = MODE_SYSLOG;
  }
}

//------------------------------------------------------------------------
void IO::setFd(FILE *fp)
{
  fpout = fp;
}

//------------------------------------------------------------------------
void IO::resetInput(void)
{
  inputsize = 0;
}

//------------------------------------------------------------------------
void IO::enableInputLimit(void)
{
  inputlimit = true;
}

//------------------------------------------------------------------------
void IO::disableInputLimit(void)
{
  inputlimit = false;
}

//------------------------------------------------------------------------
IO &IO::operator << (std::ostream &(*man)(std::ostream &))
{ 
  if (!enabled) return *this;
  
  if (useLogPrefix && outputBuffer.getSize() == 0) {
    outputBuffer << setw(5) << setfill('0') << pid;
    outputBuffer << " ";
    outputBuffer << setw(5) << setfill('0') << seqnr++;
    if (logprefix != "")
      outputBuffer << " [" << logprefix << "]";
    outputBuffer << " ";
  }

  static std::ostream &(*endl_funcpt)(std::ostream&) = std::endl;
  
  if (man == endl_funcpt) {
    outputBuffer << "\r\n";
    
    if (flushesOnEndl)
      flushContent();
  } else
    outputBuffer << man;
  
  if ((int) outputBuffer.getSize() >= buffersize)
    flushContent();

  return *this;
}

//------------------------------------------------------------------------
int IO::select(int maxfd,
		     fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
		     int &timeout)
{
  if (timeout == -1) return 1;
  struct timeval t1;
  struct timeval t2;
  int res;

  if (gettimeofday(&t1, NULL) == -1) 
    throw InputException("gettimeofday failed in select");

  do {
    struct timeval t;
    t.tv_sec = timeout / 1000;
    t.tv_usec = 1000*(timeout % 1000);
    res = ::select(maxfd, readfds, writefds, exceptfds, &t);

    if (gettimeofday(&t2, NULL) == -1)
      throw InputException("gettimeofday failed in select");

    timeout = 1000*(t2.tv_sec - t1.tv_sec) + (t2.tv_usec - t1.tv_usec)/1000;

    if (timeout < 0) timeout = 0;
    if (res == 0) return 0;
    if (res > 0) return res;
  } while (res == -1 && errno == EINTR);
  return -1;
}

//------------------------------------------------------------------------
void IO::writeStr(const string s)
{
  if (s == "")
    return;
  
  if (mode == IO::MODE_PLAIN) {
    // Set transfer timeout
    alarm(transfertimeout);
    int n = fwrite(s.c_str(), 1, s.length(), fpout);
    fflush(fpout);
    alarm(0);
    
    if (n != (int) s.length())
      throw InputException("fwrite wrote too few bytes");
  } else if (mode == IO::MODE_SYSLOG) {
    // Set transfer timeout
    alarm(transfertimeout);
    syslog(LOG_INFO, "%s", s.c_str());
    alarm(0);
  } else throw InputException("invalid write mode");
}

//------------------------------------------------------------------------
int IO::readChar(int timeout, bool retry)
{
  int nextchar = 0;
  char c;
  
  if (inputBuffer.size() != 0) {
    c = inputBuffer.top();
    inputBuffer.pop();
    return c;
  }

  if (inputlimit && inputsize >= 8192)
    throw InputException("Input limit is 8192 characters.");
  
  if (mode == IO::MODE_PLAIN) {
    fd_set rfds;
    FD_ZERO(&rfds);
    FD_SET(0, &rfds);

    int r = select(1, &rfds, NULL, NULL, timeout);
    if (r > 0) {
      if ((nextchar = read(fileno(stdin), &c, 1)) == 1) {
	nextchar = c;
	c = nextchar;
      } else
	throw InputException("readChar EOF");
    }
    else if (r == 0)
      throw TimeoutException("Reading from stdin timed out.");
    else
      throw InputException("Error reading from stdin");
  }
  
  ++inputsize;

  if (nextchar == EOF)
    throw InputException("readChar EOF");
  else
    return c;
}

//------------------------------------------------------------------------
void IO::unReadChar(int c_in) 
{
  inputBuffer.push(c_in);
}

//------------------------------------------------------------------------
void IO::unReadChar(const string &s_in)
{
  if (s_in.length() != 0) {
    int c = s_in.length() - 1;
    while (c >= 0) {
      unReadChar(s_in[c]);
      --c;
    }
  }
}

//------------------------------------------------------------------------
int IO::pending(void) const
{
  return 0;
}

//------------------------------------------------------------------------
IOFactory::IOFactory(void)
{
}

//------------------------------------------------------------------------
IOFactory::~IOFactory(void)
{
}

//------------------------------------------------------------------------
IOFactory &IOFactory::getInstance(void)
{
  static IOFactory iofactory;
  return iofactory;
}

//------------------------------------------------------------------------
void IOFactory::assign(int id, IO *io)
{
  if (ioers.find(id) != ioers.end())
    throw CriticalException("attempt to assign new IO object to existing id");

  ioers[id] = io;
}

//------------------------------------------------------------------------
IO &IOFactory::get(int id)
{
  if (ioers.find(id) == ioers.end())
    throw CriticalException("attempt to fetch IO with non-existing id");

  return *ioers[id];
}
