/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/mailbox/mailbox.h
 *  
 *  Description:
 *    Declaration of the Mailbox class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef mailbox_h_included
#define mailbox_h_included

#include <exception>
#include <map>
#include <string>
#include <queue>
#include <vector>

#include <time.h>
#include <stdio.h>
#include <string.h>

#include "message.h"
#include "status.h"
#include "../parsers/imap/imapparser.h"

namespace Binc {

  //------------------------------------------------------------------------
  class File {

    //--
    FILE *fp;

    //--
    Message & message;

    //--
    unsigned int *ref;

  public:

    //----------------------------------------------------------------------
    class const_char_crlf_iterator {

    public:

      //--
      const char operator *(void) const;

      //--
      void operator ++(void);

      //--
      void operator --(void);

      //--
      const_char_crlf_iterator(void);
      ~const_char_crlf_iterator(void);
    };

    // Constructor
    File(Message & m, FILE *f);

    // Copy constructor
    File(const File &f);

    // Conversion to FILE*
    operator FILE*() const;

    // Assignment operator
    File & operator =(const File &f);

    
    // Destructor -- closes file when all references are gone
    ~File(void);
  };

  //------------------------------------------------------------------------
  class PendingUpdates {

    //--
    std::vector<unsigned int> expunges;
    std::map<unsigned int, unsigned int> flagupdates;

    //--
    unsigned int exists;
    unsigned int recent;
    
    //--
    bool newexists;
    bool newrecent;

  public:
    enum {
      EXPUNGE = 0x01,
      FLAGS   = 0x02,
      EXISTS  = 0x04,
      RECENT  = 0x08
    };

    //----------------------------------------------------------------------
    class expunged_const_iterator {

      //--
      std::vector<unsigned int>::iterator internal;

    public:

      //--
      unsigned int operator * (void) const;

      //--
      void operator ++ (void);

      //--
      bool operator != (expunged_const_iterator) const;

      //--
      bool operator == (expunged_const_iterator) const;

      //--
      expunged_const_iterator(void);
      expunged_const_iterator(std::vector<unsigned int>::iterator i);
    };

    //--
    expunged_const_iterator beginExpunged(void);

    //--
    expunged_const_iterator endExpunged(void);

    //----------------------------------------------------------------------
    class flagupdates_const_iterator {

      //--
      std::map<unsigned int, unsigned int>::iterator internal;

    public:

      //--
      unsigned int first(void) const;

      //--
      unsigned int second(void) const;

      //--
      void operator ++ (void);

      //--
      bool operator != (flagupdates_const_iterator) const;

      //--
      flagupdates_const_iterator(void);

      //--
      flagupdates_const_iterator(std::map<unsigned int,
				 unsigned int>::iterator i);
    };

    //--
    flagupdates_const_iterator beginFlagUpdates(void);

    //--
    flagupdates_const_iterator endFlagUpdates(void);

    //--
    void addExpunged(unsigned int uid);

    //--
    void addFlagUpdates(unsigned int uid, unsigned int flags);

    //--
    void setExists(unsigned int n);

    //--
    void setRecent(unsigned int n);

    //--
    unsigned int getExists(void) const;

    //--
    unsigned int getRecent(void) const;

    //--
    bool newExists(void) const;

    //--
    bool newRecent(void) const;

    //--
    PendingUpdates();
    ~PendingUpdates();
  };

  //------------------------------------------------------------------------
  class Mailbox {
    mutable std::map<std::string, Message> messages;
  protected:

    //--
    mutable bool readonly;
    unsigned int cacheSize;
    mutable std::map<int,File*> cached;
    mutable std::queue<int> cachedQ;
    void addToCache(Message &message) const;

    //--
    mutable unsigned int oldrecent;
    mutable unsigned int oldexists;

    bool has(const std::string &id) const;
    Message & get(const std::string &id) const;
    void add(const std::string &key, Message & m) const;
    void erase(Message & m) const;

  public:
    static const int INCLUDE_EXPUNGED = 0;
    static const int SKIP_EXPUNGED    = 1;
    static const int UID_MODE         = 0;
    static const int SQNR_MODE        = 2;

    //----------------------------------------------------------------------
    class iterator {

      //--
      const Mailbox & mailbox;
      std::map<std::string, Message>::iterator internalIterator;
      Binc::BincImapParserSet bset;

      //--
      int modifier;
      unsigned int sqnr;
      int last;
      bool currentErased;

      // Private constructor; only Mailbox should be able to
      // create the iterators. We allow copying them, though.
      iterator(const Mailbox & d,
	       std::map<std::string, Message>::iterator i,
	       const Binc::BincImapParserSet &bs,
	       int modifier);

      //--
      friend class Mailbox;

      //--
      void reposition(void);

    public:

      // Erase current message
      Message & erase(void);

      // Get sqnr

      int getSqnr(void) const {
	return sqnr; 
      }

      const std::string &fixedFilename(void) {
	return internalIterator->first;
      }

      // Get message
      Message &operator * (void);

      // Equality operators
      bool operator == (const iterator &msg) const;
      bool operator != (const iterator &msg) const;

      // Advance to next message
      iterator & operator ++ (void);

      // Destructor
      ~iterator(void);
    };

    //--
    static Binc::BincImapParserSet all;

    //--
    void flushCache(void);

    //--
    const File & getFileForMessage(Message &message);

    //--
    void parseFull(Message &message) const;

    //--
    void parseOnlyHeader(Message &message) const;

    //--
    virtual bool isReadOnly(void) const
    {
      return readonly;
    }

    //--
    virtual const std::string getName(void) const = 0;

    //--
    virtual int getMaxUid(void) const;
    
    //--
    virtual int getMaxSqnr(void) const;

    //--
    virtual const std::string &getUidValidity(void) const = 0;

    //--
    virtual int getUidNext(void) const = 0;

    //--
    virtual bool getStatus(const std::string &, Status &) const
    throw (StorageException &) = 0;

    //--
    virtual int open(const std::string &s) = 0;

    //--
    virtual FILE *fopen(const std::string &s) const = 0;

    //--
    virtual PendingUpdates getUpdates(bool scan, int type) = 0;

    //--
    iterator begin(const ::Binc::BincImapParserSet 
		   &bset = Mailbox::all,
		   int mod = INCLUDE_EXPUNGED) const;

    //--
    iterator end(void) const;

    //--
    void setCacheSize(unsigned int size);

    //--
    virtual void setReadOnly(void) const
    {
      readonly = true;
    }

    //--
    virtual void bumpUidValidity(const std::string &) const = 0;

    //--
    virtual bool isMailbox(const std::string &) const = 0;

    //--
    virtual bool isMarked(const std::string &) const = 0;

    //--
    virtual int getStatusID(const std::string &) const = 0;

    //--
    virtual void scan(void) const = 0;

    //--
    virtual void imapExpunge(void) = 0;

    //--
    virtual void imapSelect(const std::string &s_in) = 0;

    //--
    virtual void imapCreate(const std::string &s, mode_t mode,
			    uid_t owner = 0, gid_t group = 0,
			    bool root = false) = 0;

    //--
    virtual void imapDelete(const std::string &s) = 0;

    //--
    virtual void imapClose(void);

    //--
    virtual void updateFlags(void) = 0;

    //--
    Mailbox(void);
    virtual ~Mailbox(void);

    //--
    friend class iterator;
  };
}

#endif
