/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/mailbox/maildir.cc
 *  
 *  Description:
 *    Implementation of the Maildir class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "maildir.h"

#include "../../io/io.h"
#include "../../util/convert.h"
#include "../../parsers/storage/storage.h"

using namespace ::std;

//------------------------------------------------------------------------
bool Binc::Maildir::readCache(void) const
{
  IO &logger = IOFactory::getInstance().get(2);

  const string uidvalfilename = path + "/bincimap-uidvalidity";
  const string cachefilename = path + "/bincimap-cache";

  bool uidvalfiledropped = false;
  bool cacheisnew = false;
  try {
    Storage uidvalfile;
    uidvalfile.load(uidvalfilename);

    uidvalfile.setSection("depot");
    uidvalidity = atoi(uidvalfile["_uidvalidity"]);
    uidnext = atoi(uidvalfile["_uidnext"]);

    if (uidvalfile["_version"] != UIDVALFILEVERSION) {
      logger << "dropping uidvalfile & cache - they're out of date." << endl;
      uidnext = 1;
      uidvalidity = time(NULL);
      uidvalfiledropped = true;
      cacheisnew = true;
    }
  } catch (StorageException &e) {
    logger << "Failed to read " << uidvalfilename << ": " 
	   << e.getReason() << endl;
    uidnext = 1;
    uidvalidity = time(NULL);
    cacheisnew = true;
  }

  try {
    Storage cache;
    cache.setSection("depot");
    cache.load(cachefilename);

    cache.setSection("depot");
    if (cache["_version"] != CACHEFILEVERSION) {
      logger << "warning: dropping cache - it's out of date (old v" 
	     << cache["_version"] << ", new v" << CACHEFILEVERSION 
	     << ")" << endl;
      if (!uidvalfiledropped) {
	uidnext = 1;
	uidvalidity = time(NULL);
	uidvalfiledropped = true;
      }
    }

    if (!uidvalfiledropped) {
      string section;
      if (cache.getFirstSection(section))
	do {
	 if (isdigit(section[0])) {
	    
	    cache.setSection(section);

	    Message m;
	    m.setID(cache["_ID"]);
	    
	    if (!has(m.getID())) {
	      m.setUID(atoi(cache["_UID"]));
	      m.setStdFlag(Message::F_JUSTARRIVED);
	      int c = atoi(cache["_Size"]);
	      if (c != 0) m.setSize(c);
	      else m.setSize(-1);
	      add(m.getID(), m);
	    }
	  }
	} while (cache.getNextSection(section));
    }
  } catch (StorageException &e) {
    logger << "warning, while reading " << cachefilename << ": " 
	   << e.getReason() << endl;
  }

  if (readonly) return !cacheisnew;

  try {
    Storage uidvalfile;
    uidvalfile.setSection("depot");
    uidvalfile.insert("_uidvalidity", toString(uidvalidity));
    uidvalfile.insert("_uidnext", toString(uidnext));
    uidvalfile.insert("_version", UIDVALFILEVERSION);
    uidvalfile.save(uidvalfilename);
  } catch (StorageException &e) {
    logger << "warning, while saving to " << uidvalfilename << ": " 
	   << e.getReason() << endl;
    throw e;
  } 

  return !cacheisnew;
}

