/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/mailbox/maildir-scan.cc
 *  
 *  Description:
 *    Implementation of the Maildir class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "maildir.h"

#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../../exceptions.h"
#include "../../io/io.h"
#include "../../parsers/storage/storage.h"

using namespace Binc;
using namespace ::std;

namespace {

  //----------------------------------------------------------------------
  class Lock {
    string lock;

  public:

    //--
    Lock(const string &path)
    {
      IO &logger = IOFactory::getInstance().get(2);

      lock = (path == "" ? "." : path) + "/bincimap-scan-lock";

      int lockfd = -1;
      while ((lockfd = ::open(lock.c_str(),
			      O_CREAT | O_WRONLY | O_EXCL)) == -1) {
	if (errno != EEXIST) {
	  logger << "Unable to aquire mailbox scan lock " << lock
		 << ", " << string(strerror(errno)) << endl;
	  throw MessageException("invalid maildir");
	}
				 
	struct stat mystat;
	logger << "Waiting for mailbox scan lock " << lock << "." << endl;
	if (lstat(lock.c_str(), &mystat) == 0) {
	  if ((time(NULL) - mystat.st_ctime) > 300) {
	    logger << "Mailbox lock " << lock << " is old. Attempting hijack."
		   << endl;

	    if (unlink(lock.c_str()) == 0) continue;
	    else logger << "Failed to hijack old mailbox scan lock " << lock
			<< ", " << string(strerror(errno)) << endl;
	  }
	} else {
	  if (errno != ENOENT) {
	    string err = "Unable to identify type of lock " + lock + ": "
	      + strerror(errno);
	    logger << err << endl;
	    throw MessageException(err);
	  }
	}
	
	// sleep one second.
	sleep(1);
      }

      close(lockfd);
    }

    //--
    ~Lock()
    {
      IO &logger = IOFactory::getInstance().get(2);

      // remove the lock
      if (unlink(lock.c_str()) != 0)
	logger << "Failed to release mailbox scan lock " << lock << ", "
	       << strerror(errno) << endl;
    }
  };
}

//------------------------------------------------------------------------
// scan the maildir. update flags, find messages in new/ and move them
// to cur, setting the recent flag in memory only. check for expunged
// messages. give newly arrived messages uids.
//------------------------------------------------------------------------
void Binc::Maildir::scan(void) const
{
  IO &logger = IOFactory::getInstance().get(2);

  static time_t old_cur_st_mtime;
  static time_t old_cur_st_ctime;
  static time_t old_new_st_mtime;
  static time_t old_new_st_ctime;

  const string newpath = path + "/new/";
  const string curpath = path + "/cur/";
  const string uidvalfilename = path + "/bincimap-uidvalidity";
  const string cachefilename = path + "/bincimap-cache";

  // lock the directory as we are scanning. this prevents race
  // conditions with uid delegation
  Lock lock(path);

  // check wether or not we need to bother scanning the folder.
  if (firstscan) {
    struct stat oldstat;
    if (stat(newpath.c_str(), &oldstat) != 0)
      throw MessageException("invalid maildir: " + string(strerror(errno)));

    old_new_st_mtime = oldstat.st_mtime;
    old_new_st_ctime = oldstat.st_ctime;

    if (stat(curpath.c_str(), &oldstat) != 0)
      throw MessageException("invalid maildir: " + string(strerror(errno)));

    old_cur_st_mtime = oldstat.st_mtime;
    old_cur_st_ctime = oldstat.st_ctime;

  } else {
    struct stat oldcurstat;
    struct stat oldnewstat;
    if (stat(newpath.c_str(), &oldnewstat) != 0)
      throw MessageException("invalid maildir: " + string(strerror(errno)));

    if (stat(curpath.c_str(), &oldcurstat) != 0)
      throw MessageException("invalid maildir: " + string(strerror(errno)));

    if (oldnewstat.st_mtime == old_new_st_mtime
	&& oldnewstat.st_ctime == old_new_st_ctime
	&& oldcurstat.st_mtime == old_cur_st_mtime
	&& oldcurstat.st_ctime == old_cur_st_ctime)
      return;

    old_cur_st_mtime = oldcurstat.st_mtime;
    old_cur_st_ctime = oldcurstat.st_ctime;
    old_new_st_mtime = oldnewstat.st_mtime;
    old_new_st_ctime = oldnewstat.st_ctime;
  }

  // If the cache files have not been read, then read them
  readCache();

  // check the uidvalidity - if another instance has bounced this
  // version, then we quit immediately.
  try {
    Storage uidvalfile;
    uidvalfile.load(uidvalfilename);
    uidvalfile.setSection("depot");
    if (uidvalfile["_version"] != UIDVALFILEVERSION) {
      logger << "uidval file v" << uidvalfile["_version"]
	     << " too old (new is v" << UIDVALFILEVERSION << ")";

      unlink(cachefilename.c_str());
      unlink(uidvalfilename.c_str());

      throw StorageException("uidvalidity bounced - exiting!");
    }
  } catch (StorageException &e) {
    logger << "warning while loading " << uidvalfilename << ": "
	   << e.getReason() << endl;

    throw e;
  }

  bool uidnextchanged = false;
  bool mailboxchanged = false;

  // open directory
  DIR *pdir = opendir(newpath.c_str());
  if (pdir == NULL) {
    string reason = "Maildir::scan::opendir(\"" + newpath + "\") == NULL (";
    reason += strerror(errno);
    reason += ")";

    throw MessageException(reason);
  }
  
  // scan all entries
  struct dirent *pdirent;
  while ((pdirent = readdir(pdir)) != NULL) {

    // skip files whose filename does not start with a digit.
    if (!isdigit(pdirent->d_name[0]))
      continue;

    string filename = pdirent->d_name;
    string time;
    string hostname;
    string nrepeat_count;
    string info;

    string::size_type pos = filename.find('.');

    // skip entry that does not have a '.' in it.
    if (pos == string::npos)
      continue;

    // extract part 1, the time_t
    time = filename.substr(0, pos);

    // this is important. do not move messages from new/ that are not
    // at least one second old or messages may disappear. this
    // introduces a special case: we can not cache the old st_ctime
    // and st_mtime 
    if (::time(NULL) <= atoi(time.c_str())) {
      old_cur_st_mtime = (time_t) 0;
      old_cur_st_ctime = (time_t) 0;
      old_new_st_mtime = (time_t) 0;
      old_new_st_ctime = (time_t) 0;

      logger << "Skipping message that is less that one second old: " << filename << endl;
      continue;
    }

    filename = filename.substr(pos + 1);
    pos = filename.find('.');

    // extract the non-repeating section.
    nrepeat_count = filename.substr(0, pos);

    // extract the host name. we now have all parts that uniquely
    // identify this message. what comes after this may vary.
    hostname = filename.substr(pos + 1);    

    BincStream ss;
     
    ss << setw(10) << setfill('0') << atoi(time.c_str());

    string fixedfilename = ss.str();
    fixedfilename += "." + nrepeat_count + "." + hostname;
    
    // move files from new/ to cur/
    if (link((newpath + fixedfilename).c_str(), 
	       (curpath + fixedfilename).c_str()) != 0) {
      // FIXME: this error condition is actually quite serious, and
      // should be handled by creating a new file name I guess.
      logger << "skipping " << newpath 
	     << pdirent->d_name << ": could not rename it" << endl;
      continue;
    } else if (unlink((newpath + fixedfilename).c_str()) != 0)
      logger << "warning: unlink(" << newpath  + fixedfilename << ") == " 
	     << errno << ": " << strerror(errno) << endl;
  }
  closedir(pdir);
  
  // Now, assume all known messages were expunged
  Mailbox::iterator e = end();
  for (Mailbox::iterator i = begin(); i != e; ++i)
    (*i).setExpunged();

  // Then, scan cur
  // open directory
  pdir = opendir(curpath.c_str());
  if (pdir == NULL) {
    string reason = "Maildir::scan::opendir(\"" + curpath + "\") == NULL (";
    reason += strerror(errno);
    reason += ")";

    throw MessageException(reason);
  }

  // erase all old maps between fixed filenames and actual file names
  fixedtofile.clear();

  // scan all entries
  while ((pdirent = readdir(pdir)) != NULL) {
    if (!isdigit(pdirent->d_name[0]))
      continue;

    string filename = pdirent->d_name;
    string time;
    string hostname;
    string nrepeat_count;
    string info;

    string::size_type pos = filename.find('.');
    if (pos == string::npos) continue;

    time = filename.substr(0, pos);
    filename = filename.substr(pos + 1);

    if ((pos = filename.find('.')) == string::npos) continue;

    nrepeat_count = filename.substr(0, pos);
    filename = filename.substr(pos + 1);    

    
    if ((pos = filename.find(':')) == string::npos)
      hostname = filename;
    else {
      hostname = filename.substr(0, pos);
      info = filename.substr(pos + 1);    
    }

    string experimental;
    string standard;
    if (info != "")
      if ((pos = info.find("2,")) != string::npos) {
	standard = info.substr(pos + 2);
      }

    Message m;

    for (string::const_iterator i = standard.begin();
	 i != standard.end(); ++i) {
      switch (*i) {
      case 'R': m.setStdFlag(Message::F_ANSWERED); break;
      case 'S': m.setStdFlag(Message::F_SEEN); break;
      case 'T': m.setStdFlag(Message::F_DELETED); break;
      case 'D': m.setStdFlag(Message::F_DRAFT); break;
      case 'F': m.setStdFlag(Message::F_FLAGGED); break;
      default:
	// error in std flags
	continue;
      }
    }

    BincStream ss;

    ss << setw(10) << setfill('0') << atoi(time.c_str());

    string fixedfilename = ss.str();
    fixedfilename += "." + nrepeat_count + "." + hostname;

    fixedtofile[fixedfilename] = pdirent->d_name;

    // if we already have details about this file in memory..
    if (has(fixedfilename)) {
      Message &message = get(fixedfilename);

      // then confirm that this message was not expunged
      message.setUnExpunged();

      // update the in-memory data with what flags we found in this
      // entry.
      if (m.getStdFlags() != (message.getStdFlags() & ~Message::F_RECENT)) {
	// the flags are different. we need to report this.
	message.resetStdFlags();
	message.setStdFlag(m.getStdFlags());
      }

      continue;
    }

    m.setSize(-1);
    m.setStdFlag(Message::F_RECENT);
    m.setID(fixedfilename);

    // for now, lets not give this message an ID.
    m.setUID(-1);
    add(fixedfilename, m);
  }

  closedir(pdir);

  // All new messages get a UID. The important this about this pass is
  // that messages are ordered internally by their fixed (padded) file
  // name.  So all messages must be in a strict ascending order, and
  // all messages with no UID must be at the end. If this is not so,
  // then we need to bounce the uidvalidity and exit.
  bool lasthaduid = true;
  Mailbox::iterator i = begin(Mailbox::all, 
				   Mailbox::INCLUDE_EXPUNGED);
  Mailbox::iterator e2 = end();
  for (; i != e2; ++i) {
    Message & message = *i;
    if (message.isExpunged()) {
      // FIXME: Must make a check here..
      mailboxchanged = true;
    }
    
    if (message.getUID() == -1) {
      message.setUID(uidnext++);
      mailboxchanged = true;
      uidnextchanged = true;
      lasthaduid = false;
    } else {
      // ok, so we found a message that already had a UID.  if the
      // last message we processed did /not/ have a UID, then we
      // bounce.
      if (!lasthaduid) {
	logger << "found a recent message with uid "
	       << message.getUID() << " _after_ one that had no uid" << endl;
	// bounce uidvalidity.
	if (!readonly) {
	  try {
	    Storage uidvalfile;
	    uidvalfile.setSection("depot");
	    uidvalfile.insert("_uidvalidity", toString(++uidvalidity));
	    uidvalfile.insert("_uidnext", toString(1));
	    uidvalfile.insert("_version", UIDVALFILEVERSION);
	    uidvalfile.save(uidvalfilename);
	  } catch (StorageException &e) {
	    logger << "warning when saving " << uidvalfilename 
		   << ": " << e.getReason() << endl;
	    
	    throw e;
	  }
	  
	  // and exit
	  unlink(cachefilename.c_str());
	  
	  // remove the uidvalidity file
	  unlink(uidvalfilename.c_str());
	  
	  throw StorageException("uidvalidity bounced");
	} else
	  throw StorageException("uidvalidity would bounce");
      }
    }
  }

  // Messages that existed in the cache that we read, but did not
  // exist in the Maildir, are removed from the messages list.
  Mailbox::iterator jj = begin(all, INCLUDE_EXPUNGED);
  Mailbox::iterator jje = end();
  while (jj != jje) {
    Message & message = *jj;
    if (message.isExpunged() && (message.getStdFlags() & Message::F_JUSTARRIVED))
      jj.erase();
    ++jj;
  }

  // Special case: The first time we scan is in SELECT. All flags
  // changes for new messages will then appear to be recent, and
  // to avoid this to be sent to the client as a pending update,
  // we explicitly unset to "flagsChanged" flag in all messages.
  if (firstscan) {
    firstscan = false;
    int lastuid = 0;

    Mailbox::iterator e = end();
    for (Mailbox::iterator i = begin(); i != e; ++i) {
      Message & message = *i;
      if (lastuid < message.getUID())
	lastuid = message.getUID();
      else {
	// remove the cache file
	unlink(cachefilename.c_str());
	
	// remove the uidvalidity file
	unlink(uidvalfilename.c_str());
	
	// FIXME: can we recover? if this is the first scan, perhaps?
	throw StorageException("uidvalidity bounced - exiting!");
      }
      
      message.setFlagsUnchanged();
    }
  }


  if (firstscan || (!readonly && mailboxchanged))
    writeCache();

  if (firstscan || (!readonly && uidnextchanged))
    try {
      Storage uidvalfile;
      uidvalfile.setSection("depot");
      uidvalfile.insert("_uidvalidity", toString(uidvalidity));
      uidvalfile.insert("_uidnext", toString(uidnext));
      uidvalfile.insert("_version", UIDVALFILEVERSION);
      uidvalfile.save(uidvalfilename);
    } catch (StorageException &e) {
      logger << "warning while saving " << uidvalfilename 
	     << ": " << e.getReason() << endl;
      throw e;
    }
}
