/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/mailbox/maildir.cc
 *  
 *  Description:
 *    Implementation of the Maildir class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <iomanip>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "../../io/io.h"
#include "../../parsers/storage/storage.h"
#include "../../util/convert.h"
#include "maildir.h"

using namespace ::std;
using namespace Binc;

//------------------------------------------------------------------------
Maildir::Maildir(void) : Mailbox()
{
  firstscan = true;
  cacheRead = false;
  uidvalidity = -1;
  uidnext = 1;
  selected = false;
  oldrecent = 0;
  oldexists = 0;
}

//------------------------------------------------------------------------
Maildir::~Maildir(void)
{
}

//------------------------------------------------------------------------
void Maildir::setPath(const string &path_in)
{
  path = path_in;
}

//------------------------------------------------------------------------
PendingUpdates Maildir::getUpdates(bool doscan, int type)
{
  PendingUpdates updates;

  if (doscan)
    scan();

  flushCache();

  int sqnr = 1;
  unsigned int exists = 0;
  unsigned int recent = 0;

  Mailbox::iterator e = end();
  
  // count messages, find recent
  if (!readonly && (type & PendingUpdates::EXPUNGE)) {
    
    Mailbox::iterator i = begin(Mailbox::all, Mailbox::INCLUDE_EXPUNGED);
    for (; i != e; ++i) {
      Message & message = *i;
      if (message.isExpunged()) {
	updates.addExpunged(sqnr);
	i.erase();
      } else ++sqnr;
    }
  }

  Mailbox::iterator i = begin(Mailbox::all, Mailbox::INCLUDE_EXPUNGED);
  while (i != e) {
    Message & message = *i;
    // at this point, there is a message that is not expunged
    ++exists;
    if (message.getStdFlags() & Message::F_RECENT) ++recent;
    ++i;
  }

  if (exists != oldexists)
    updates.setExists(oldexists = exists);

  if (recent != oldrecent)
    updates.setRecent(oldrecent = recent);
  
  if (type & PendingUpdates::FLAGS) {   
    for (Mailbox::iterator i = begin(); i != e; ++i) {
      Message & message = *i;
      if (message.hasFlagsChanged()) {
	int flags = message.getStdFlags();
	
	updates.addFlagUpdates(i.getSqnr(), flags);

	message.setFlagsUnchanged();
      }
    }
  }  

  return updates;
}

//------------------------------------------------------------------------
bool Maildir::isMailbox(const std::string &s_in) const
{
  std::string reason;
  struct stat mystat;
  if (stat(s_in.c_str(), &mystat) == 0) {
    if (S_ISDIR(mystat.st_mode)) {
      if ((stat((s_in + "/cur").c_str(), &mystat) == 0 
	   && S_ISDIR(mystat.st_mode))
	  && (stat((s_in + "/new").c_str(), &mystat) == 0 
	      && S_ISDIR(mystat.st_mode))
	  && (stat((s_in + "/tmp").c_str(), &mystat) == 0 
	      && S_ISDIR(mystat.st_mode)))
	return true;
    }
  }

  return false;
}

//------------------------------------------------------------------------
const std::string Maildir::getName(void) const
{
  return "Maildir";
}

//------------------------------------------------------------------------
void Maildir::bumpUidValidity(const string &s_in) const
{
  unlink((s_in + "/bincimap-uidvalidity").c_str());
  unlink((s_in + "/bincimap-cache").c_str());
}

//------------------------------------------------------------------------
bool Maildir::isMarked(const std::string &s_in) const
{
  DIR *dirp = opendir((s_in + "/new").c_str());
  if (dirp == NULL) return false;

  struct dirent *direntp;  
  while ((direntp = readdir(dirp)) != NULL)
    if (isdigit(direntp->d_name[0])) {
      closedir(dirp);
      return true;
    }

  closedir(dirp);
  return false;
}

//------------------------------------------------------------------------
int Maildir::getStatusID(const string &path) const 
{
  int statusid = 0;
  struct stat mystat;
  if (stat((path + "/new").c_str(), &mystat) == 0)
    statusid = mystat.st_ctime;

  if (stat((path + "/cur").c_str(), &mystat) == 0)
    statusid *= mystat.st_ctime;

  return statusid;
}

//------------------------------------------------------------------------
bool Maildir::getStatus(const string &path, Status &s) const 
  throw (StorageException &)
{
  int messages = 0;
  int unseen = 0;
  int messagesincache = 0;

  const string cachefilename = path + "/bincimap-cache";
  const string uidvalfilename = path + "/bincimap-uidvalidity";

  Storage cache;
    Storage uidvalfile;

  // Read cache and uidvalidity files
  try {
    cache.load(cachefilename);
    uidvalfile.load(uidvalfilename);
  } catch (StorageException &e) {
  }

  string section;
  if (cache.getFirstSection(section))
    do {
      if (isdigit(section[0]))
	++messagesincache;
    } while (cache.getNextSection(section));
  
  uidvalfile.setSection("depot");
  const int uidvalidity = atoi(uidvalfile["_uidvalidity"].c_str());
  const int uidnext =  atoi(uidvalfile["_uidnext"].c_str());
  
  s.setUidValidity(uidvalidity < 1 ? time(NULL) : uidvalidity);
  s.setUidNext(uidnext < 1 ? 1 : uidnext);

  // Scan new
  DIR *dirp = opendir((path + "/new").c_str());
  if (dirp == NULL) return false;

  struct dirent *direntp;  
  while ((direntp = readdir(dirp)) != NULL)
    if (isdigit(direntp->d_name[0])) {
      ++messages;
      ++unseen;
    }

  closedir(dirp);

  // Scan cur
  if ((dirp = opendir((path + "/cur").c_str())) == NULL)
    return false;

  while ((direntp = readdir(dirp)) != NULL)
    if (isdigit(direntp->d_name[0])) {
      ++messages;
      const string dname = direntp->d_name;
      const string::size_type pos = dname.find(':');
      if (pos != string::npos)
	if (dname.substr(pos).find('S') == string::npos)
	  ++unseen;
    }
  closedir(dirp);
  
  int recent = messages - messagesincache;
  s.setRecent(recent < 0 ? 0 : recent);
  s.setMessages(messages);
  s.setUnseen(unseen);

  return true;
}
