/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-copy.cc
 *  
 *  Description:
 *    Implementation of the COPY command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>

#include "depot.h"
#include "io/io.h"
#include "mailbox/maildir/maildir.h"
#include "operators.h"
#include "parsers/imap/recursivedescent/recursivedescent.h"
#include "session.h"
#include "util/convert.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
CopyOperator::CopyOperator(void)
{
}

//----------------------------------------------------------------------
CopyOperator::~CopyOperator(void)
{
}

//----------------------------------------------------------------------
const string CopyOperator::getName(void) const
{
  return "COPY";
}

//----------------------------------------------------------------------
int CopyOperator::getState(void) const
{
  return Session::SELECTED;
}

//------------------------------------------------------------------------
void CopyOperator::process(Depot &depot, Command &command) const
{
  IO &logger = IOFactory::getInstance().get(2);

  const string &srcmailbox = command.getMailbox();
  Mailbox *mailbox = NULL;

  if ((mailbox = depot.get(srcmailbox)) == NULL)
    throw OperatorException(Operator::NO, "invalid mailbox " 
			    + toImapString(srcmailbox));

  const string &canonmailbox = toCanonMailbox(srcmailbox);
  const string &maildir = depot.mailboxToFilename(canonmailbox);

  int nmessages = 0;
  int mode;
  if (command.getUidMode())
    mode = Mailbox::UID_MODE;
  else 
    mode = Mailbox::SQNR_MODE;

  Mailbox::iterator i = mailbox->begin(command.bset,
				       Mailbox::SKIP_EXPUNGED|mode);
  Mailbox::iterator e = mailbox->end();
  for (; i != e;++i) {
    Message & message = *i;

    // Create a tmp/ write filename
    string safenamestr = maildir + "/tmp/bincimap-copy-XXXXXX";
    // convert it to a c string 
    size_t safenamelength = safenamestr.length() + 1;
    /* TODO: Check this against _PC_PATH_MAX */
    // Here we take advantage of C++'s stack allocation of arrays --cld
    char *safename = new char[safenamelength];
    memset(safename, 0, safenamelength);
    strncpy(safename, safenamestr.c_str(), safenamelength-1);
    
    int fd = mkstemp(safename);
    // see if mkstemp() failed highly unlikely but not impossible --cld
    if (fd == -1) {
      logger << "COPY unable to create temp file '" << safenamestr
	     << "', unrecoverable error:" << strerror(errno) << endl;
      delete safename;
      continue;  // continue with next copy (I think) --cld
    }
    
    int fdread = mailbox->open(message.getID());
    if(fdread == -1) {
      logger << "COPY couldn't open " << message.getID() << endl;
      unlink(safename);
      delete safename;
      continue;	
    }
    
    // Query the OS to find out the best size for I/O on that file --cld
    size_t buffsz = (size_t)fpathconf(fdread, _PC_PIPE_BUF);
    
    // again, a stack buffer to ensure automatic cleanup --cld
    char *copybuffer = new char[buffsz];
    ssize_t bytesread = 0;    // the number of bytes read each cycle
    bool broke = false;       // ABEND flag from while loop --cld
    
    // At this point, fdread needs to be copied safely to fd
    while (!broke && (bytesread = read(fdread, copybuffer, buffsz)) != 0) {
      
      // if a failure occurs in this loop, remove the file, log
      // a message, set broke = true and break;
      
      if (bytesread == -1) {
	// an error occured in the read. --cld
	close(fdread);
	unlink(safename);
	delete safename;
	delete copybuffer;

	logger << "COPY failed during read() of " << message.getID() 
	       << ": " << strerror(errno) << endl;
	broke = true;
	break;
      }
      
      ssize_t byteswritten = 0;
      // attempt to write the entire contents of the buffer
      while (byteswritten != bytesread) {
	ssize_t thesebytes = 
	  write(fd, copybuffer + byteswritten, bytesread-byteswritten);
	if(thesebytes == -1) {
	  close(fdread);
	  unlink(safename);
	  logger << "COPY couldn't read " << message.getID() 
		 << ": " << strerror(errno) << endl;
	  broke = true;
	  break;
	} else {
	  byteswritten += thesebytes;
	}
      }
    }
    
    delete copybuffer;

    close(fdread);
    if (broke) continue;
    
    if (fsync(fd) != 0) {
      if (errno != EINVAL) { // EINVAL simply means the FS doesn't do sync
	// recovery
	logger << "COPY fsync failed " << strerror(errno) << endl;
	int err = errno;
	close(fd);
	unlink(safename);
	delete safename;
	throw OperatorException(Operator::NO, strerror(err));
      }
    }
    
    // at this point, the tmp/ file has been created, now we deliver --cld
    
    char hostname[512];
    int hostnamelen = gethostname(hostname, sizeof(hostname));
    if (hostnamelen == -1 || hostnamelen == sizeof(hostname))
      strcpy(hostname, "localhost");
    
    int fileid = time(NULL);
    string filename;
    
    if (close(fd) != 0) {
      int err = errno;
      unlink(safename);
      delete safename;
      logger << "COPY couldn't close or link file: " << (char*)safename
	     << ": " << strerror(err) << endl;
      throw OperatorException(Operator::NO, strerror(err));
    }
    
    bool copied = false;
    // loop trying to deliver into new/
    for (int i = 0; i < 1000; ++i) {
      BincStream ss;

      if (i == 0)
	ss << maildir << "/new/" << fileid 
	   << "." << getpid() << "." << hostname;
      else
	ss << maildir << "/new/" << fileid 
	   << "." << getpid() << "_" << i << "." << hostname;
      
      filename = ss.str();
      
      // if link fails, we _can_ and _should_ try to create another file name
      if (link(safename, filename.c_str()) == 0) {
	nmessages++;
	copied = true;
	break;
      } else if (errno != EEXIST)
	break;
    }
    
    if (!copied) {
      int err = errno;
      unlink(safename);
      logger << "COPY couldn't close or link file: " << (char*)safename
	     << ": " << strerror(err) << endl;
      delete safename;
      throw OperatorException(Operator::NO, strerror(err));
    }
    
    if (unlink(safename) != 0)
      logger << "COPY warning: couldn't unlink " << (char*)safename
	     << ": " << strerror(errno) << endl;
    delete safename;
  }
}

//------------------------------------------------------------------------
bool CopyOperator::parse(Command &c_in) const
{
  try {
    bool uidmode = false;
    if (expectThisString("UID")) {
      uidmode = true;
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
    }

    if (!expectThisString("COPY")) {
      IO &com = IOFactory::getInstance().get(1);

      if (uidmode)
	com.unReadChar("UID ");
      return false;
    }

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    if (!expectSet(c_in.getSet()))
      throw SyntaxException("expected set");

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    string mailbox;
    if (!expectMailbox(mailbox))
      throw SyntaxException("expected mailbox");
    if (!expectCRLF())
      throw SyntaxException("expected CRLF");

    if (uidmode) c_in.setUidMode();
    c_in.setMailbox(mailbox);
    c_in.setName("COPY");
    return true;
  } catch (InputException &e) {
    e.addReason("in CopyOperator::parse");
    throw e;
  }
  return false;
}
