/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-login.cc
 *  
 *  Description:
 *    Implementation of the LOGIN command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>

#include <string>
#include <iostream>

#include "parsers/imap/recursivedescent/recursivedescent.h"
#include "session.h"
#include "authenticate.h"

#include "io/io.h"

#include "depot.h"
#include "operators.h"

using namespace ::std;
using namespace Binc;


//----------------------------------------------------------------------
LoginOperator::LoginOperator(void)
{
}

//----------------------------------------------------------------------
LoginOperator::~LoginOperator(void)
{
}

//----------------------------------------------------------------------
const string LoginOperator::getName(void) const
{
  return "LOGIN";
}

//----------------------------------------------------------------------
int LoginOperator::getState(void) const
{
  return Session::NONAUTHENTICATED;
}

//------------------------------------------------------------------------
void LoginOperator::process(Depot &depot, Command &command) const
{
  Session &session = Session::getInstance();
  IO &logger = IOFactory::getInstance().get(2);

  session.globalconfig.setSection("Authentication");
  const bool allowplain = (session.globalconfig["allow plain auth in non ssl"] == "yes");

  if (!session.command.ssl && !allowplain)
    throw OperatorException(Operator::NO, "disallowed in plain-text connection");

  putenv(strdup(("BINCIMAP_LOGIN=LOGIN+" + command.getTag()).c_str()));

  switch (authenticate(depot, command.getUserID(), command.getPassword())) {
  case 1: 
    throw OperatorException(Operator::NO, "internal error");
    return;
  case 2:
    throw OperatorException(Operator::NO, "wrong user ID or password");
  }

  // go to logout
  session.setState(Session::LOGOUT);
  throw OperatorException(Operator::NOTHING, "");
}

//----------------------------------------------------------------------
bool LoginOperator::parse(Command &c_in) const
{
  try {
    if (!expectThisString("LOGIN"))
      return false;
    
    if (!expectSPACE())
      throw SyntaxException("expected single SPACE after LOGIN");

    string userid;
    if (!expectAstring(userid))
      throw SyntaxException("expected userid after LOGIN SPACE");

    c_in.setUserID(userid);

    if (!expectSPACE())
      throw SyntaxException("expected SPACE after LOGIN SPACE userid");

    string password;
    if (!expectAstring(password))
      throw SyntaxException("expected password after LOGIN SPACE userid SPACE");
    if (!expectCRLF())
      throw SyntaxException("expected CRLF after password");

    c_in.setPassword(password);

  } catch (InputException &e) {
    e.addReason("in LoginOperator::parse");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in LoginOperator::parse");
    throw e;
  }

  c_in.setName("LOGIN");
  return true;
}
