/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/parsers/imap/imapparser.cc
 *  
 *  Description:
 *    Implementation of the common items for parsing IMAP input
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "imapparser.h"
#include "../../util/convert.h"

#include <stdio.h>
#include <map>
#include <iostream>
#include <vector>
#include <string>
#include <exception>

using namespace ::std;
using namespace Binc;

//------------------------------------------------------------------------
Command::Command(void) 
  : extra(NULL), flags(), statuses(), bset(), searchkey(), fatt()
{
  uidmode = false;
}

Command::~Command(void)
{
  if (extra != 0)
    delete extra;
}

//------------------------------------------------------------------------
void Command::setUidMode(void)
{
  uidmode = true;
}

//------------------------------------------------------------------------
bool Command::getUidMode(void) const
{
  return uidmode;
}

//------------------------------------------------------------------------
void Command::setTag(string &t_in)
{
  tag = t_in;
}

//------------------------------------------------------------------------
const string &Command::getTag(void) const
{
  return tag;
}

//------------------------------------------------------------------------
void Command::setMode(const string &m_in)
{
  mode = m_in;
}

//------------------------------------------------------------------------
const string &Command::getMode(void) const
{ 
  return mode;
}

//------------------------------------------------------------------------
void Command::setName(const string &s_in)
{
  name = s_in;
}

//------------------------------------------------------------------------
const string &Command::getName(void) const
{
  return name;
}

//------------------------------------------------------------------------
void Command::setAuthType(const string &s_in)
{
  authtype = s_in;
}

//------------------------------------------------------------------------
const string &Command::getAuthType(void) const
{
  return authtype;
}

//------------------------------------------------------------------------
void Command::setLiteral(const string &s_in)
{
  literal = s_in;
}

//------------------------------------------------------------------------
const string &Command::getLiteral(void) const
{
  return literal;
}

//------------------------------------------------------------------------
void Command::setDate(const string &s_in)
{
  date = s_in;
}

//------------------------------------------------------------------------
const string &Command::getDate(void) const
{
  return date;
}

//------------------------------------------------------------------------
void Command::setCharSet(const string &s_in)
{
  charset = s_in;
}

//------------------------------------------------------------------------
const string &Command::getCharSet(void) const
{
  return charset;
}

//------------------------------------------------------------------------
void Command::setUserID(const string &s_in)
{
  userid = s_in;
}

//------------------------------------------------------------------------
const string &Command::getUserID(void) const
{
  return userid;
}

//------------------------------------------------------------------------
void Command::setPassword(const string &s_in)
{
  password = s_in;
}

//------------------------------------------------------------------------
const string &Command::getPassword(void) const
{
  return password;
}

//------------------------------------------------------------------------
void Command::setMailbox(const string &s_in)
{
  mailbox = s_in;
}

//------------------------------------------------------------------------
const string &Command::getMailbox(void) const
{
  return mailbox;
}

//------------------------------------------------------------------------
void Command::setListMailbox(const string &s_in)
{
  listmailbox = s_in;
}

//------------------------------------------------------------------------
const string &Command::getListMailbox(void) const
{
  return listmailbox;
}

//------------------------------------------------------------------------
void Command::setNewMailbox(const string &s_in)
{
  newmailbox = s_in;
}

//------------------------------------------------------------------------
const string &Command::getNewMailbox(void) const
{
  return newmailbox;
}

//------------------------------------------------------------------------
BincImapParserSet &Command::getSet(void)
{  
  return bset;
}

//------------------------------------------------------------------------
vector<string> &Command::getStatuses(void)
{
  return statuses;
}

//------------------------------------------------------------------------
vector<string> &Command::getFlags(void)
{
  return flags;
}

//------------------------------------------------------------------------
BincImapParserSet::BincImapParserSet(void)
{
  limited = true;
}

//------------------------------------------------------------------------
BincImapParserSet::~BincImapParserSet(void)
{
}

//------------------------------------------------------------------------
BincImapParserSet::Range::Range(unsigned int a, unsigned int b)
{
  if (a > b) {
    from = b;
    to = a;
  } else {
    from = a;
    to = b;
  }
}

//------------------------------------------------------------------------
void BincImapParserSet::addRange(unsigned int a, unsigned int b)
{
  if (a == (unsigned int)-1 || b == (unsigned int)-1) limited = false;
  internal.push_back(Range(a, b));
}

//------------------------------------------------------------------------
void BincImapParserSet::addNumber(unsigned int a)
{
  if (a == (unsigned int)-1) limited = false;
  internal.push_back(Range(a, a));
}

//------------------------------------------------------------------------
bool BincImapParserSet::isInSet(unsigned int n) const
{
  for (vector<Range>::const_iterator i = internal.begin();
       i != internal.end(); ++i) {
    if (n >= (*i).from && n <= (*i).to)
      return true;
  }

  return false;
}

//------------------------------------------------------------------------
BincImapParserFetchAtt::BincImapParserFetchAtt(void)
{
  offsetstart = 0;
  offsetlength = -1;
  hassection = false;
}

//------------------------------------------------------------------------
string BincImapParserFetchAtt::toString(void)
{
  string tmp;
  if (type == "BODY.PEEK")
    tmp = "BODY";
  else
    tmp = type;

  if (type == "BODY" || type == "BODY.PEEK") {
    if (hassection) {
      tmp += "[";
      tmp += section;
      if (sectiontext != "") {
	if (section != "")
	  tmp += ".";
	tmp += sectiontext;
	
	if (headerlist.size() != 0) {
	  tmp += " (";
	  for (vector<string>::iterator i = headerlist.begin();
	       i != headerlist.end(); ++i) {
	    if (i != headerlist.begin())
	      tmp += " ";
	    tmp += *i;
	  }
	  tmp += ")";
	}
      }
      tmp += "]";
    
      if (offsetstart == 0 && offsetlength == -1)
	tmp += " ";
      else
	tmp += "<" + Binc::toString(offsetstart) + "> ";
    }
  }

  return tmp;
}

//------------------------------------------------------------------------
BincImapParserSearchKey::BincImapParserSearchKey(void)
{
  type = 0;
  number = 0;
}

//------------------------------------------------------------------------
const BincImapParserSet& BincImapParserSearchKey::getSet(void) const
{
  return bset;
}
