/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/parsers/imap/recursivedescent/recursivedescent.cc
 *  
 *  Description:
 *    Implementation of a recursive descent IMAP command
 *    parser.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// #define DEBUG

#include "../imapparser.h"
#include "recursivedescent.h"
#include "../../../io/io.h"
#include "../../../util/convert.h"

#include <stdio.h>
#include <ctype.h>
#include <stack>
#include <iostream>
#include <iomanip>

using namespace ::std;

stack<int> Binc::inputBuffer;
int Binc::charnr = 0;


//----------------------------------------------------------------------
bool Binc::expectThisString(const string &s_in)
{
  IO &com = IOFactory::getInstance().get(1);

#ifdef DEBUG
  cout << "expectThisString(\"" << s_in << "\")" << endl << flush;
#endif
  string tmp;

  try {
    bool match = true;
    for (string::const_iterator i = s_in.begin(); i != s_in.end(); ++i) {

      int c = com.readChar();
      tmp += c;

      if (toupper(*i) != toupper(c)) {
	match = false;
	break;
      }
    }

    if (!match) {
      com.unReadChar(tmp);
      return false;
    } else
      return true;

  } catch (InputException &e) {
    e.addReason("in expectThisString");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectDateTime(string &s_in)
{
  try {
    if (!expectThisString("\""))
      return false;

    int digit1, digit2;
    if (expectSPACE()) {
      digit1 = 0;
      if (!expectDigit(digit2))
	throw SyntaxException("expected digit (day) after \" and a SPACE.");
    } else {
      if (!expectDigit(digit1))
	throw SyntaxException("expected first digit of day");
      if (!expectDigit(digit2))
	throw SyntaxException("expected second digit of day");
    }

    int day = digit1 * 10 + digit2;

    BincStream daystr;

    daystr << setw(2) << setfill('0') << day;

    s_in += daystr.str();

    if (!expectThisString("-"))
      throw SyntaxException("expected -");
    s_in += "-";

    /* month */
    if (expectThisString("Jan")) s_in += "Jan";
    else if (expectThisString("Feb")) s_in += "Feb";
    else if (expectThisString("Mar")) s_in += "Mar";
    else if (expectThisString("Apr")) s_in += "Apr";
    else if (expectThisString("May")) s_in += "May";
    else if (expectThisString("Jun")) s_in += "Jun";
    else if (expectThisString("Jul")) s_in += "Jul";
    else if (expectThisString("Aug")) s_in += "Aug";
    else if (expectThisString("Sep")) s_in += "Sep";
    else if (expectThisString("Oct")) s_in += "Oct";
    else if (expectThisString("Nov")) s_in += "Nov";
    else if (expectThisString("Dec")) s_in += "Dec";
    else
      throw SyntaxException("expected month");

    if (!expectThisString("-"))
      throw SyntaxException("expected -");
    s_in += "-";    

    /* year */
    int year, c;
    if (!expectDigit(year)) throw SyntaxException("expected digit (first digit of year)");
    if (!expectDigit(c)) throw SyntaxException("expected digit (second digit of year)");
    year = (year * 10) + c;
    if (!expectDigit(c)) throw SyntaxException("expected digit (third digit of year)");
    year = (year * 10) + c;
    if (!expectDigit(c)) throw SyntaxException("expected digit (last digit of year)");
    year = (year * 10) + c;

    BincStream yearstr;
    
    yearstr << year;

    s_in += yearstr.str();

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    s_in += " ";

    if (!expectTime(s_in))
      throw SyntaxException("expected time");

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    s_in += " ";

    if (!expectZone(s_in))
      throw SyntaxException("expected zone");

    if (!expectThisString("\""))
      throw SyntaxException("expected \"");

    return true;

  } catch (InputException &e) {
    e.addReason("in expectDateTime");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectDateTime");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectTime(string &s_in)
{
  try {
    int c, t;
    if (!expectDigit(t))
      return false;
    if (!expectDigit(c))
      throw SyntaxException("expected digit");
    t = (t * 10) + c;

    BincStream tstr;

    tstr << t;

    s_in += tstr.str();

    if (!expectThisString(":"))
      throw SyntaxException("expected colon");

    s_in += ":";

    if (!expectDigit(t))
      throw SyntaxException("expected digit");
    if (!expectDigit(c))
      throw SyntaxException("expected digit");
    t = (t * 10) + c;

    tstr.clear();

    tstr << t;

    s_in += tstr.str();

    if (!expectThisString(":"))
      throw SyntaxException("expected colon");

    s_in += ":";

    if (!expectDigit(t))
      throw SyntaxException("expected digit");
    if (!expectDigit(c))
      throw SyntaxException("expected digit");
    t = (t * 10) + c;

    tstr.clear();

    tstr << t;

    s_in += tstr.str();

    return true;

  } catch (InputException &e) {
    e.addReason("in expectTime");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectTime");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectZone(string &s_in)
{
  try {
    if (expectThisString("-"))
      s_in += "-";
    else if (expectThisString("+"))
      s_in += "+";
    else
      return false;

    int c, t;
    if (!expectDigit(t))
      throw SyntaxException("expected digit");
    if (!expectDigit(c))
      throw SyntaxException("expected digit");
    t = (t * 10) + c;
    if (!expectDigit(c))
      throw SyntaxException("expected digit");
    t = (t * 10) + c;
    if (!expectDigit(c))
      throw SyntaxException("expected digit");
    t = (t * 10) + c;

    BincStream tstr;

    tstr << t;

    s_in += tstr.str();
    
    return true;

  } catch (InputException &e) {
    e.addReason("in expectZone");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectZone");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectListWildcards(int &c_in)
{
  try {
    if (expectThisString("%")) {
      c_in = '%';
      return true;
    } else if (expectThisString("*")) {
      c_in = '*';
      return true;
    } else
      return false;
  } catch (InputException &e) {
    e.addReason("in expectListMailbox");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectListMailbox");
    throw e;
  } 
}

//----------------------------------------------------------------------
bool Binc::expectListMailbox(string &s_in)
{
  IO &com = IOFactory::getInstance().get(1);

  try {
    if (expectString(s_in))
      return true;
    
    int c;
    if (expectAtomChar(c) || expectListWildcards(c) || expectThisString("]"))
      do {
        s_in += (char) c;
        if (!expectAtomChar(c) && !expectListWildcards(c) && !expectThisString("]"))
	  return true;
      } while (1);
    
    com.unReadChar(s_in);
    
    return false;
  } catch (InputException &e) {
    e.addReason("in expectListMailbox");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectListMailbox");
    throw e;
  }

  return true;
}



//----------------------------------------------------------------------
bool Binc::expectFlag(vector<string> &v_in)
{
  try {
    string flag;
    if (expectThisString("\\Answered")) v_in.push_back("\\Answered");
    else if (expectThisString("\\Flagged")) v_in.push_back("\\Flagged");
    else if (expectThisString("\\Deleted")) v_in.push_back("\\Deleted");
    else if (expectThisString("\\Seen")) v_in.push_back("\\Seen");
    else if (expectThisString("\\Draft")) v_in.push_back("\\Draft");
    else if (expectThisString("\\Answered")) v_in.push_back("\\Answered");
    else {
      if (expectThisString("\\")) {
	if (expectAtom(flag))
	  v_in.push_back("\\" + flag);
	else
	  throw SyntaxException("expected atom");
      } else if (expectAtom(flag)) {
	v_in.push_back(flag);
      } else
	return false;
    }

    return true;

  } catch (InputException &e) {
    e.addReason("in expectFlag");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectFlag");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectDate(string &s_in)
{
  try {
    bool quoted = false;
    if (expectThisString("\""))
      quoted = true;

    /* day */
    int day, c;
    if (expectDigit(c)) {
      day = c;
      if (expectDigit(c))
	day = (day * 10) + c;

      BincStream daystr;

      daystr << day;

      s_in += daystr.str();
    } else
      throw SyntaxException("expected digit");

    /* - */
    if (!expectThisString("-"))
      throw SyntaxException("expected -");
    s_in += '-';

    /* month */
    if (expectThisString("Jan")) s_in += "Jan";
    else if (expectThisString("Feb")) s_in += "Feb";
    else if (expectThisString("Mar")) s_in += "Mar";
    else if (expectThisString("Apr")) s_in += "Apr";
    else if (expectThisString("May")) s_in += "May";
    else if (expectThisString("Jun")) s_in += "Jun";
    else if (expectThisString("Jul")) s_in += "Jul";
    else if (expectThisString("Aug")) s_in += "Aug";
    else if (expectThisString("Sep")) s_in += "Sep";
    else if (expectThisString("Oct")) s_in += "Oct";
    else if (expectThisString("Nov")) s_in += "Nov";
    else if (expectThisString("Dec")) s_in += "Dec";
    else
      throw SyntaxException("expected month");
      
    /* - */
    if (!expectThisString("-"))
      throw SyntaxException("expected -");
    s_in += '-';

    /* year */
    int year;
    if (!expectDigit(year)) throw SyntaxException("expected digit");
    if (!expectDigit(c)) throw SyntaxException("expected digit");
    year = (year * 10) + c;
    if (!expectDigit(c)) throw SyntaxException("expected digit");
    year = (year * 10) + c;
    if (!expectDigit(c)) throw SyntaxException("expected digit");
    year = (year * 10) + c;

    BincStream yearstr;

    yearstr << year;

    s_in += yearstr.str();

    if (quoted)
      if (!expectThisString("\""))
	throw SyntaxException("expected \"");
    
    return true;

  } catch (InputException &e) {
    e.addReason("in expectDate");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectDate");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectCRLF(void)
{
  try {
    if (expectCR() && expectLF())
      return true;
    else
      return false;
  } catch (InputException &e) {
    e.addReason("in expectCRLF");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectCR(void)
{
  IO &com = IOFactory::getInstance().get(1);

  try {
    int c = com.readChar();
    return (c == 0x0d);

  } catch (InputException &e) {
    e.addReason("in expectCR");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectLF(void)
{
  IO &com = IOFactory::getInstance().get(1);

  try {
    int c = com.readChar();
    return (c == 0x0a);

  } catch (InputException &e) {
    e.addReason("in expectLF");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectTagChar(int &c_in)
{
  IO &com = IOFactory::getInstance().get(1);

  int c;

  try {
    c = com.readChar();
  } catch (InputException &e) {
    e.addReason("in expectTagChar");
    throw e;
  }

  switch (c) {
    case 041:    case 043:    case 044:    case 046:    case 047:    case 054:
    case 055:    case 056:    case 057:    case 060:    case 061:    case 062:
    case 063:    case 064:    case 065:    case 066:    case 067:    case 070:
    case 071:    case 072:    case 073:    case 074:    case 075:    case 076:
    case 077:    case 0100:   case 0101:   case 0102:   case 0103:   case 0104:
    case 0105:   case 0106:   case 0107:   case 0110:   case 0111:   case 0112:
    case 0113:   case 0114:   case 0115:   case 0116:   case 0117:   case 0120:
    case 0121:   case 0122:   case 0123:   case 0124:   case 0125:   case 0126:
    case 0127:   case 0130:   case 0131:   case 0132:   case 0133:   case 0135:
    case 0136:   case 0137:   case 0140:   case 0141:   case 0142:   case 0143:
    case 0144:   case 0145:   case 0146:   case 0147:   case 0150:   case 0151:
    case 0152:   case 0153:   case 0154:   case 0155:   case 0156:   case 0157:
    case 0160:   case 0161:   case 0162:   case 0163:   case 0164:   case 0165:
    case 0166:   case 0167:   case 0170:   case 0171:   case 0172:   case 0174:
    case 0175:   case 0176:
      c_in = c;
      return true;
  default:
    break;
  }

  try {
    com.unReadChar(c);
  } catch (InputException &e) {
    e.addReason("in expectTagChar");
    throw e;
  }

  return false;
}

//----------------------------------------------------------------------
bool Binc::expectTag(string &s_in)
{
  IO &com = IOFactory::getInstance().get(1);

  string tag;
  int tagchar;

  try {
    while (expectTagChar(tagchar))
      tag += tagchar;

    if (tag == "") {
      com.unReadChar(tag);
      return false;
    } else
      s_in = tag;

  } catch (InputException &e) {
    e.addReason("in expectTag");
    throw e;
  }

  return true;
}

//----------------------------------------------------------------------
bool Binc::expectSPACE(void)
{
  IO &com = IOFactory::getInstance().get(1);

  try {
    int c = com.readChar();
    if (c == ' ') return true;

    com.unReadChar(c);
  } catch (InputException &e) {
    e.addReason("in expectSPACE");
    throw e;
  }

  return false;
}

//----------------------------------------------------------------------
bool Binc::expectMailbox(string &s_in)
{
  try {
    if (expectAstring(s_in))
      return true;
  } catch (InputException &e) {
    e.addReason("in expectMailbox");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectMailbox");
    throw e;
  }

  return false;
}

//----------------------------------------------------------------------
bool Binc::expectAstring(string &s_in)
{
  try {
    if (expectAtom(s_in))
      return true;
    
    if (expectString(s_in))
      return true;
  } catch (InputException &e) {
    e.addReason("in expectAstring");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectAstring");
    throw e;
  }

  return false;
}

//----------------------------------------------------------------------
bool Binc::expectAtomChar(int &c_in)
{
  IO &com = IOFactory::getInstance().get(1);

  int c;

  try {
    c = com.readChar();
  } catch (InputException &e) {
    e.addReason("in expectAtomChar");
    throw e;
  }

  switch (c) {
  case 041:    case 043:    case 044:    case 046:    case 047:    case 053:
  case 054:    case 055:    case 056:    case 057:    case 060:    case 061:
  case 062:    case 063:    case 064:    case 065:    case 066:    case 067:
  case 070:    case 071:    case 072:    case 073:    case 074:    case 075:
  case 076:    case 077:    case 0100:   case 0101:   case 0102:   case 0103:
  case 0104:   case 0105:   case 0106:   case 0107:   case 0110:   case 0111:
  case 0112:   case 0113:   case 0114:   case 0115:   case 0116:   case 0117:
  case 0120:   case 0121:   case 0122:   case 0123:   case 0124:   case 0125:
  case 0126:   case 0127:   case 0130:   case 0131:   case 0132:   case 0133:
  case 0135:   case 0136:   case 0137:   case 0140:   case 0141:   case 0142:
  case 0143:   case 0144:   case 0145:   case 0146:   case 0147:   case 0150:
  case 0151:   case 0152:   case 0153:   case 0154:   case 0155:   case 0156:
  case 0157:   case 0160:   case 0161:   case 0162:   case 0163:   case 0164:
  case 0165:   case 0166:   case 0167:   case 0170:   case 0171:   case 0172:
  case 0174:   case 0175:   case 0176:
    c_in = c;
    return true;
  default:
    break;
  }
  
  try {
    com.unReadChar(c);
  } catch (InputException &e) {
    e.addReason("in expectAtomChar");
    throw e;
  }

  return false;
}

//----------------------------------------------------------------------
bool Binc::expectAtom(string &s_in)
{
  IO &com = IOFactory::getInstance().get(1);

  string atom;
  int atomchar;

  try {
    while (expectAtomChar(atomchar))
      atom += atomchar;

    if (atom == "") {
      com.unReadChar(atom);
      return false;
    } else
      s_in = atom;

  } catch (InputException &e) {
    e.addReason("in expectAtom");
    throw e;
  }

  return true;
}

//----------------------------------------------------------------------
bool Binc::expectString(string &s_in)
{
  if (expectQuoted(s_in))
    return true;

  if (expectLiteral(s_in))
    return true;

  return false;
}

//----------------------------------------------------------------------
bool Binc::expectQuoted(string &s_in)
{
  string quoted;
  int quotedchar;

  try {
    if (!expectThisString("\""))
      return false;
    
    while (expectQuotedChar(quotedchar))
      quoted += quotedchar;
    
    if (!expectThisString("\""))
      return false;
  } catch (InputException &e) {
    e.addReason("in expectQuoted");
    throw e;
  }

  s_in = quoted;

  return true;
}

//----------------------------------------------------------------------
bool Binc::expectQuotedChar(int &c_in)
{
  IO &com = IOFactory::getInstance().get(1);

  int c;

  try {
    c = com.readChar();
  } catch (InputException &e) {
    e.addReason("in expectQuotedChar");
    throw e;
  }

  switch (c) {
  case 01:    case 02:    case 03:    case 04:    case 05:    case 06:    case 07:
  case 010:   case 011:   case 013:   case 014:   case 016:   case 017:
  case 020:   case 021:   case 022:   case 023:   case 024:   case 025:   case 026:   case 027:
  case 030:   case 031:   case 032:   case 033:   case 034:   case 035:   case 036:   case 037:
  case 040:   case 041:   case 043:   case 044:   case 045:   case 046:   case 047:
  case 050:   case 051:   case 052:   case 053:   case 054:   case 055:   case 056:   case 057:
  case 060:   case 061:   case 062:   case 063:   case 064:   case 065:   case 066:   case 067:
  case 070:   case 071:   case 072:   case 073:   case 074:   case 075:   case 076:   case 077:
  case 0100:  case 0101:  case 0102:  case 0103:  case 0104:  case 0105:  case 0106:  case 0107:
  case 0110:  case 0111:  case 0112:  case 0113:  case 0114:  case 0115:  case 0116:  case 0117:
  case 0120:  case 0121:  case 0122:  case 0123:  case 0124:  case 0125:  case 0126:  case 0127:
  case 0130:  case 0131:  case 0132:  case 0133:  case 0135:  case 0136:  case 0137:
  case 0140:  case 0141:  case 0142:  case 0143:  case 0144:  case 0145:  case 0146:  case 0147:
  case 0150:  case 0151:  case 0152:  case 0153:  case 0154:  case 0155:  case 0156:  case 0157:
  case 0160:  case 0161:  case 0162:  case 0163:  case 0164:  case 0165:  case 0166:  case 0167:
  case 0170:  case 0171:  case 0172:  case 0173:  case 0174:  case 0175:  case 0176:  case 0177:
    c_in = c;
    return true;
  case '\\': {
    int d;
    try {
      d = com.readChar();
    } catch (InputException &e) {
      e.addReason("in expectQuotedChar");
      throw e;
    }

    if (d == '\"' || d == '\\') {
      c_in = d;
      return true;
    } else {
      try {
	com.unReadChar(d);
	com.unReadChar(c);
      } catch (InputException &e) {
	e.addReason("in expectQuotedChar");
	throw e;
      }
    }
  }
  default:
    break;
  }
  
  try {
    com.unReadChar(c);
  } catch (InputException &e) {
    e.addReason("in expectQuotedChar");
    throw e;
  }

  return false;
  

}

//----------------------------------------------------------------------
bool Binc::expectLiteral(string &s_in)
{
  IO &com = IOFactory::getInstance().get(1);

  string literal;

  try {
    if (!expectThisString("{"))
      return false;

    int nchar;
    if (!expectNumber(nchar))
      throw SyntaxException("expected number");

    if (!expectThisString("}"))
      throw SyntaxException("expected }");

    if (!expectCRLF())
      throw SyntaxException("expected CRLF");

    com << "+ ok, send " << nchar << " bytes of data." << endl;
    com.flushContent();

    for (int i = 0; i < nchar; ++i)
      literal += com.readChar();

  } catch (InputException &e) {
    e.addReason("in expectLiteral");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectLiteral");
    throw e;
  }

  s_in = literal;

  return true;
}

//----------------------------------------------------------------------
bool Binc::expectNumber(int &i_in)
{
  try {
    i_in = 0;
    int n;
    while (expectDigit(n)) {
      if (i_in == 0)
	i_in = n;
      else
	i_in = (i_in * 10) + n;
    }

    return true;

  } catch (InputException &e) {
    e.addReason("in expectNumber");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectDigit(int &i_in)
{
  IO &com = IOFactory::getInstance().get(1);

  try {  
    int c = com.readChar();
    if (c == '0') {
      i_in = 0;
      return true;
    } else
      com.unReadChar(c);

    if (!expectDigitNZ(i_in))
      return false;

  } catch (InputException &e) {
    e.addReason("in expectNumber");
    throw e;
  }

  return true;
}

//----------------------------------------------------------------------
bool Binc::expectDigitNZ(int &i_in)
{
  IO &com = IOFactory::getInstance().get(1);

  try {  
    int c;
    switch ((c = com.readChar())) {
    case '1': i_in = 1; break;
    case '2': i_in = 2; break;
    case '3': i_in = 3; break;
    case '4': i_in = 4; break;
    case '5': i_in = 5; break;
    case '6': i_in = 6; break;
    case '7': i_in = 7; break;
    case '8': i_in = 8; break;
    case '9': i_in = 9; break;
    default:
      com.unReadChar(c);
      return false;
    }

    return true;

  } catch (InputException &e) {
    e.addReason("in expectNumber");
    throw e;
  }

  return true;
}

//----------------------------------------------------------------------
bool Binc::expectSet(BincImapParserSet &s_in)
{
  try {
    int seqnum = -1;

    /* if a set does not start with a sequencenum, then it's not a
     * set. :-) seqnum == -1 means '*'. */
    if (!expectSequenceNum(seqnum))
      return false;
    
    /* the first number is always a part of the set */
    s_in.addNumber(seqnum);

    /* if _after_ a set there is a ':', then there will always be a
     * sequencenum after the colon. if not, it's a syntax error. a
     * colon delimits two numbers in a range. */
    if (expectThisString(":")) {
      int seqnum2 = -1;
      if (!expectSequenceNum(seqnum2))
	throw SyntaxException("expected sequencenum");

      s_in.addRange(seqnum, seqnum2);
    }

    /* if _after_ a set there is a ',', then there will always be
     * a set after the comma. if not, it's a syntax error. */
    if (expectThisString(","))
      if (!expectSet(s_in))
	throw SyntaxException("expected set");

  } catch (InputException &e) {
    e.addReason("in expectSet");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectSet");
    throw e;
  }

  return true;
}

//----------------------------------------------------------------------
bool Binc::expectSequenceNum(int &i_in)
{
  IO &com = IOFactory::getInstance().get(1);

  try {
    int c = com.readChar();
    if (c == '*') {
      i_in = -1;
      return true;
    } else
      com.unReadChar(c);

    if (!expectNZNumber(i_in))
      return false;
    else
      return true;
  } catch (InputException &e) {
    e.addReason("in expectSet");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectSet");
    throw e;
  }
}

//----------------------------------------------------------------------
bool Binc::expectNZNumber(int &i_in)
{
  try {
    int c;
    if (!expectDigitNZ(c))
      return false;
    
    i_in = c;
    while (expectDigit(c))
      i_in = (i_in * 10) + c;

  } catch (InputException &e) {
    e.addReason("in expectNZNumber");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectNZNumber");
    throw e;
  }

  return true;
}
